-- -------------------
-- STRUKTUR VON ORACLE
-- -------------------

-- Informationen ber Schemaobjekte

SELECT tablespace_name, block_size, status, logging FROM USER_TABLESPACES;
SELECT table_name, tablespace_name FROM USER_TABLES;
SELECT view_name, text FROM USER_VIEWS;
SELECT index_name, Index_type, table_name, uniqueness FROM USER_INDEXES;
SELECT constraint_name, constraint_type, table_name FROM USER_CONSTRAINTS;

-- -----------------------------
-- BESONDERHEITEN VON ORACLE SQL
-- -----------------------------

-- RowID

SELECT rowid, gkz, name FROM Gemeinden;


-- BLOBs

CREATE TABLE Rasterkarten (
  id    INTEGER,                     -- ID
  name  VARCHAR(80)  NOT NULL,       -- Kartenname
  karte BLOB,                        -- Karte
  CONSTRAINT pk_rk PRIMARY KEY (id)  -- Primrschlssel
)
LOB (karte) STORE AS (
  CHUNK 16K
);

ALTER TABLE Rasterkarten MODIFY LOB (karte) (CACHE); 


-- BFILEs

CREATE TABLE GemeindeVideos (
  id     INTEGER,                     -- ID
  gem    DECIMAL(8),                  -- Gemeindeschlssel
  video  BFILE,                       -- Video von der Gemeinde
  CONSTRAINT pk_gv PRIMARY KEY (id),  -- Primrschlssel
  CONSTRAINT fk_gv_gem FOREIGN KEY (gem)
    REFERENCES Gemeinden(gkz) ON DELETE CASCADE  -- Fremdschlssel
);

CREATE OR REPLACE DIRECTORY VideoOrdner AS 'D:/videos';

INSERT INTO GemeindeVideos VALUES (1, 3403000, BFILENAME('VideoOrdner','oldenburg_video.avi'));
COMMIT;


-- Sequenzgeneratoren

-- Sequenzgenerator anlegen:
CREATE SEQUENCE Hochschulen_Seq START WITH 4;
-- Sequenzgenerator verwenden:
INSERT INTO Hochschulen (num,name)
                        VALUES (Hochschulen_Seq.nextval,'Uni Bremen');
COMMIT;

-- Ergebnis berprfen:
SELECT * FROM Hochschulen WHERE Name = 'Uni Bremen';
-- aktuellen Wert des Sequenzgenerators abfragen:
SELECT Hochschulen_Seq.currval FROM DUAL;

-- Metadaten abfragen:
SELECT * FROM USER_SEQUENCES;

-- ------------------------------
-- PROGRAMMIERUNG UND DATENIMPORT
-- ------------------------------

-- PL/SQL

-- Funktion anlegen:
CREATE OR REPLACE FUNCTION kreisstadttext (name VARCHAR) RETURN VARCHAR
IS
BEGIN
  IF name IS NULL THEN
    RETURN 'keine Kreisstadt';
  ELSE
    RETURN name;
  END IF;
END;
/

-- Funktion aufrufen:
SELECT Gemeinden.name, kreisstadttext(Kreise.name)
FROM Gemeinden LEFT JOIN Kreise
ON gkz = kreisstadt;

--- Log-Tabelle anlegen:
CREATE TABLE LogHochschulen (
  hs     INTEGER,     -- Hochschulnummer
  stud   DECIMAL(6),  -- Anzahl Studierende
  datum  TIMESTAMP    -- Einfgezeitpunkt
);

--- PL/SQL-Block ausfhren:
DECLARE
  -- Deklaration Cursor und Variable
  CURSOR c IS SELECT * FROM Hochschulen;   -- DB-Cursor
  akt_t c%ROWTYPE;                 -- aktuelle Antwort
BEGIN
  -- alle Antworten durchlaufen
  FOR akt_t IN c LOOP
    INSERT INTO LogHochschulen VALUES (akt_t.num, akt_t.stud, sysdate);
  END LOOP;
END;
/

--- Tabelle ausgeben:
SELECT * FROM LogHochschulen;

-- Alternativ mit benannter PL/SQL-Prozedur:

-- Datenstze lschen:
ROLLBACK;

--- Benannte Prozedur deklarieren:
CREATE OR REPLACE PROCEDURE fuelleLogHS IS
  -- Deklaration Cursor und Variable
  CURSOR c IS SELECT * FROM Hochschulen;   -- DB-Cursor
  akt_ds c%ROWTYPE;                 -- aktuelle Antwort
BEGIN
  -- alle Antworten durchlaufen
  FOR akt_ds IN c LOOP
    -- Datensatz einfgen
    INSERT INTO LogHochschulen VALUES (akt_ds.num,akt_ds.stud,sysdate);
  END LOOP;
END;
/

-- Prozedur aufrufen, Commit:
EXECUTE fuelleLogHS();
COMMIT;

SELECT * FROM LogHochschulen;

-- Sperren von Datenstzen beim Update:
DECLARE
  studNum  DECIMAL(6);   -- Anzahl der Studenten
BEGIN
  SELECT stud INTO studNum FROM Hochschulen WHERE num = 3 FOR UPDATE;
    -- nun ist der Datensatz gesperrt, so dass eine komplexe Aktion folgen knnte
  studNum := studNum - 500;
  UPDATE Hochschulen SET stud = studNum WHERE num = 3;
  COMMIT;
END;
/


-- Trigger

-- Einfge-Trigger anlegen:
CREATE OR REPLACE TRIGGER insert_hs_trg BEFORE INSERT ON Hochschulen
FOR EACH ROW
BEGIN
  INSERT INTO LogHochschulen VALUES (:new.num,:new.stud,sysdate);
END;
/

-- Update-Trigger anlegen:
CREATE OR REPLACE TRIGGER update_hs_trg BEFORE UPDATE ON Hochschulen
FOR EACH ROW
BEGIN
  IF :old.stud <> :new.stud THEN
    INSERT INTO LogHochschulen VALUES (:new.num,:new.stud,sysdate);
  END IF;
END;
/

-- Lsch-Trigger anlegen:
CREATE OR REPLACE TRIGGER delete_hs_trg BEFORE DELETE ON Hochschulen
FOR EACH ROW
BEGIN
  INSERT INTO LogHochschulen VALUES (:old.num,0,sysdate);
END;
/

-- nderungen in der Tabelle Hochschulen durchfhren, um Trigger auszulsen:
INSERT INTO Hochschulen VALUES (99,'LMU Mnchen',50000);
UPDATE Hochschulen SET stud = 55000 WHERE num = 99;
DELETE FROM Hochschulen WHERE num = 99;

COMMIT;

--- Aktionen des Triggers berprfen:
SELECT stud, datum
FROM LogHochschulen
WHERE hs = 99
ORDER BY datum;


-- Java-Routine

-- Funktion anlegen:
CREATE OR REPLACE FUNCTION kreisstadttext_jv (name VARCHAR)
                                              RETURN VARCHAR IS
  LANGUAGE JAVA
        NAME 'Kreise.gibText (java.lang.String) return java.lang.String';
/

-- Von KOMMANDOZEILE aus laden:
loadjava -user <Benutzer>/<Passwort>@<Dienstname> -andresolve Kreise.java

-- Funktion aufrufen:
SELECT Gemeinden.name, kreisstadttext_jv(Kreise.name)
FROM Gemeinden LEFT JOIN Kreise
ON gkz = kreisstadt;


-- SQL*Loader

-- Aufruf des SQL*Loaders von der KOMMANDOZEILE aus:
sqlldr USERID=<Benutzer>@<Dienstname> CONTROL=standorte.ctl LOG=standorte.log ERRORS=50
sqlldr USERID=<Benutzer>@<Dienstname> CONTROL=kreise.ctl LOG=kreise.log ERRORS=50

-- Daten berprfen:
SELECT * FROM HatStandortIn;
SELECT * FROM Kreise;


-- -----------------------------------------
-- OBJEKTRELATIONALE ERWEITERUNGEN IN ORACLE
-- -----------------------------------------

-- Felder

--- Definition:
CREATE OR REPLACE TYPE TextFeld AS VARRAY(5) OF VARCHAR(20);
/

--- Erweiterung der Tabelle "GemeindeVideos":
ALTER TABLE GemeindeVideos ADD inhalt TextFeld;

UPDATE GemeindeVideos
SET inhalt = TextFeld('Marktplatz','Schloss','Lange Strae')
WHERE gem = 3403000;

COMMIT;

--- Nutzung:
SELECT id, gem, inhalt FROM GemeindeVideos WHERE inhalt IS NOT NULL;

-- Zugriff auf Feldelemente:

CREATE OR REPLACE FUNCTION elementAusTextFeld (i DECIMAL, feld TextFeld) RETURN VARCHAR
IS
BEGIN
  IF feld IS NULL THEN
    RETURN '/';
  ELSIF (i >= 1) AND (i <= feld.count) THEN
    RETURN feld(i);
  ELSE
    RETURN '-';
  END IF;
END;
/

SELECT id, gem
FROM GemeindeVideos
WHERE elementAusTextFeld(2,inhalt) = 'Schloss';

-- einen zustzlichen Eintrag anfgen (Lsung mit Cursor):
DECLARE
  CURSOR c IS SELECT inhalt FROM GemeindeVideos WHERE inhalt IS NOT NULL FOR UPDATE;   -- DB-Cursor
  akt_t c%ROWTYPE;                 -- aktuelle Antwort
BEGIN
  FOR akt_t IN c LOOP
    IF akt_t.inhalt.count < akt_t.inhalt.limit THEN
      akt_t.inhalt.extend(1);
      akt_t.inhalt(akt_t.inhalt.count) := 'Bahnhof';
      UPDATE GemeindeVideos SET inhalt = akt_t.inhalt WHERE CURRENT OF c;
    END IF;
  END LOOP;
  COMMIT;
END;
/

-- einen zustzlichen Eintrag anfgen (alternative Lsung fr einen Datensatz):
DECLARE
  feld  TextFeld;   -- das Feld
BEGIN
  SELECT inhalt INTO feld FROM GemeindeVideos WHERE id = 1 FOR UPDATE;
  feld.extend(1);
  feld(feld.count) := 'Weser-Ems-Halle';
  UPDATE GemeindeVideos SET inhalt = feld WHERE id = 1;
  COMMIT;
END;
/

-- Ergebnis anzeigen:
SELECT inhalt FROM GemeindeVideos WHERE inhalt IS NOT NULL;

-- Klassen

-- Definition des Datentyps "Punkt2D_Typ"

CREATE OR REPLACE TYPE Punkt2D_Typ AS OBJECT (
  x  DECIMAL(10,7),       -- x-Koordinate in Grad
  y  DECIMAL(9,7)         -- y-Koordinate in Grad
) NOT FINAL;
/

-- Erweiterung der Tabelle Gemeinden

-- Attribut hinzufgen:
ALTER TABLE Gemeinden ADD centrum Punkt2D_Typ;

-- Daten einfgen:
UPDATE Gemeinden SET centrum=Punkt2D_Typ(9.7358,52.3803) WHERE gkz = 3241001;
UPDATE Gemeinden SET centrum=Punkt2D_Typ(8.2275,53.1375) WHERE gkz = 3403000;
UPDATE Gemeinden SET centrum=Punkt2D_Typ(7.9981,53.1844) WHERE gkz = 3451002;
UPDATE Gemeinden SET centrum=Punkt2D_Typ(7.9256,53.2581) WHERE gkz = 3451007;
UPDATE Gemeinden SET centrum=Punkt2D_Typ(7.2072,53.3686) WHERE gkz = 3402000;
UPDATE Gemeinden SET centrum=Punkt2D_Typ(8.0450,52.2810) WHERE gkz = 3404000;
UPDATE Gemeinden SET centrum=Punkt2D_Typ(8.1081,53.5397) WHERE gkz = 3405000;
COMMIT;

-- Anzeige:
SELECT gkz, name, centrum FROM Gemeinden;

-- Untersttzung von Methoden

--- nderung der Spezifikation:
ALTER TYPE Punkt2D_Typ REPLACE AS OBJECT (
  x  DECIMAL(10,7),       -- x-Koordinate in Grad
  y  DECIMAL(9,7),        -- y-Koordinate in Grad
  --
  MEMBER FUNCTION istWestlichVon (vgl Punkt2D_Typ) RETURN INTEGER,
  -- Testet, ob das Punktobjekt westlich vom Punkt vgl liegt.
  -- Falls ja, wird 1, sonst 0 zurckgegeben.
  --
  MEMBER FUNCTION gebeBeschreibung RETURN VARCHAR,
  -- Gibt vom Objekt eine Beschreibung als String zurck.
  --
  STATIC FUNCTION nullpkt RETURN Punkt2D_Typ
  -- Gibt einen Nullpunkt zurck.
) NOT FINAL;

--- Implementierung:
CREATE OR REPLACE TYPE BODY Punkt2D_Typ AS
  MEMBER FUNCTION istWestlichVon (vgl Punkt2D_Typ) RETURN INTEGER IS
    BEGIN
      IF x < vgl.x THEN  -- entspricht: IF SELF.x < vgl.x THEN
        RETURN 1;
      ELSE
        RETURN 0;
      END IF;
    END;
  MEMBER FUNCTION gebeBeschreibung RETURN VARCHAR IS
    BEGIN
      RETURN '(' || x || ' , ' || y || ')';
    END;
  STATIC FUNCTION nullpkt RETURN Punkt2D_Typ IS
    BEGIN
      RETURN Punkt2D_Typ(0,0);
    END;
END;
/

--- Nutzung:
SELECT name, g.centrum.gebeBeschreibung() FROM Gemeinden g;

SELECT g2.name, g2.centrum.gebeBeschreibung()
FROM Gemeinden g1 INNER JOIN Gemeinden g2
ON g1.gkz = 3403000 
WHERE g2.centrum.istWestlichVon(g1.centrum) = 1;

SELECT name
FROM Gemeinden g
WHERE g.centrum.istWestlichVon(Punkt2D_Typ.nullpkt()) = 0;

-- Objekttabellen

-- Anlegen:
CREATE TABLE Messpunkte OF Punkt2D_Typ;

-- Einfgen:
INSERT INTO Messpunkte VALUES (9.6351,52.3799);
INSERT INTO Messpunkte VALUES (9.6410,52.3813);
INSERT INTO Messpunkte VALUES (9.6359,52.3795);
COMMIT;

-- Attributwerte abfragen:
SELECT * FROM Messpunkte;

-- Objekte abfragen:
SELECT VALUE(m) FROM Messpunkte m;
-- ab Version 10 auch:
SELECT object_value FROM Messpunkte;
-- auch mglich:
SELECT SYS_NC_ROWINFO$ FROM Messpunkte;

-- OID abfragen:
SELECT object_id FROM Messpunkte;
-- auch mglich:
SELECT SYS_NC_OID$ FROM Messpunkte;

-- Objektreferenz REF abfragen:
SELECT REF(m) FROM Messpunkte m;

-- Tabelle mit Referenztyp

-- Tabelle anlegen:
CREATE TABLE Messungen (
  datum       DATE,                                -- Datumsangabe
  temperatur  DECIMAL(5,2),                        -- Temperaturangabe
  ort         REF Punkt2D_Typ SCOPE IS Messpunkte  -- Referenz auf Messpunkte
);

-- Datenstze einfgen:
INSERT INTO Messungen
VALUES ('07.12.2004',7.3,
        (SELECT REF(m) FROM Messpunkte m WHERE m.x = 9.6351) );
COMMIT;

-- Auflsen von Referenzen
SELECT datum, temperatur, DEREF(ort) FROM Messungen;

SELECT datum, temperatur, m.ort.x, m.ort.y FROM Messungen m;


-- Vererbung

--- Spezifikation:
CREATE TYPE BenannterPunkt2D_Typ UNDER Punkt2D_Typ (
  name VARCHAR(30),             -- der Name
  --  
  OVERRIDING MEMBER FUNCTION gebeBeschreibung RETURN VARCHAR,
  -- Gibt vom Objekt eine Beschreibung als String zurck.
  --
  MEMBER FUNCTION hatGleichenNamen (p BenannterPunkt2D_Typ) RETURN INTEGER
  -- Testet, ob das Punktobjekt den gleichen Namen wie p hat.
  -- Falls ja, wird 1, sonst 0 zurckgegeben.
) NOT FINAL;
/

--- Implementierung:
CREATE OR REPLACE TYPE BODY BenannterPunkt2D_Typ AS
  OVERRIDING MEMBER FUNCTION gebeBeschreibung RETURN VARCHAR IS
    BEGIN
      RETURN name || ': ' || '(' || x || ' , ' || y || ')' ;
    END;
  MEMBER FUNCTION hatGleichenNamen (p BenannterPunkt2D_Typ) RETURN INTEGER IS
    BEGIN
      IF name = p.name THEN
        RETURN 1;
      ELSE
        RETURN 0;
      END IF;
    END;
END;
/

--- Nutzung:
INSERT INTO Gemeinden (gkz,name,einw,centrum) VALUES (4011000,'Bremen',540950,BenannterPunkt2D_Typ(8.8139,53.0839,'Marktplatz'));

UPDATE Gemeinden SET centrum=BenannterPunkt2D_Typ(8.2275,53.1375,'Schlossplatz') WHERE gkz = 3403000;
UPDATE Gemeinden SET centrum=BenannterPunkt2D_Typ(7.9256,53.2581,'Marktplatz') WHERE gkz = 3451007;

COMMIT;

SELECT name, g.centrum.gebeBeschreibung() FROM Gemeinden g;

-- Casting:
SELECT g1.name, g2.name, TREAT(g1.centrum AS BenannterPunkt2D_Typ).hatGleichenNamen(TREAT(g2.centrum AS BenannterPunkt2D_Typ))
FROM Gemeinden g1 INNER JOIN Gemeinden g2
ON g1.gkz < g2.gkz
WHERE g1.centrum IS OF(BenannterPunkt2D_Typ) AND g2.centrum IS OF(BenannterPunkt2D_Typ);
